#include "mytime.h"
#include <ESP8266WiFi.h>
#include <ESP8266HTTPClient.h>
#include <WiFiUdp.h>
#include "LiquidCrystal_I2C.h"

// hodiny
mytime myclock;

// ntp klient
unsigned int localPort = 2390;
IPAddress timeServerIP;
const char* ntpServerName = "cz.pool.ntp.org";
const int NTP_PACKET_SIZE = 48;
byte packetBuffer[ NTP_PACKET_SIZE];
WiFiUDP udp;
#define STBY  0
#define SYNC  1
#define READP 2
int ntpState = SYNC;

// serial
#define RXBUFSIZE 200
#define TMPBUFSIZE 10
#define ITEMSCOUNT 8
unsigned char rxBuffer[RXBUFSIZE];
char rxTmp[TMPBUFSIZE];
unsigned char rxPtr = 0, rxCislo = 0, rxPtrn = 0;
unsigned int rxData[ITEMSCOUNT];

// wifi
const char* ssid = "mojewifi";
const char* password = "heslohesloxx";
int failedWiFi = 0;
unsigned char httpFails = 0;

// GPIO vstupy/vystupy
#define FVESWITCHENABLEPIN D5
#define FVERESTARTPIN D6

unsigned char fveSwitchEnable = 0;

// struktura na data z MPPT regulatoru
typedef struct  {
  unsigned int dataage = 0;
  unsigned int enable = 0;
  unsigned int p = 0;
  unsigned int u = 0;
  unsigned int i = 0;
  unsigned int wh = 0;
  unsigned int pwm = 0;
} FV;

// pole na ulozeni dat s MPPT regulatoru
FV fvdata[1];

// nastaveni LCD 20x4
LiquidCrystal_I2C lcd(0x27, 20, 4);




// ----------------------------------------------------------------------------------------------
// SETUP funkce
// ----------------------------------------------------------------------------------------------
void setup() {
  int maxRetries;

  // -- vypnuti psa --
  ESP.wdtDisable();

  // -- povoleni/zakazani prepnuti na FVE (zbytecne prepinani v noci) --
  pinMode(FVESWITCHENABLEPIN, INPUT);
  digitalWrite(FVESWITCHENABLEPIN, LOW);

  // -- provede restart MPPT tak, aby se smazala denni vyroba v noci --
  pinMode(FVERESTARTPIN, INPUT);
  digitalWrite(FVERESTARTPIN, LOW);

  // -- interni LED --
  pinMode(LED_BUILTIN, OUTPUT);

  // -- aktivace seriove linky --
  Serial.begin(115200);
  Serial.println();

  // -- aktivace UDP stacku pro komunikaci s NTP serverem --
  udp.begin(localPort);

  // -- nastaveni LCD 20x4 displeje a uvodni obrazovka --
  lcd.begin();
  lcd.print("PSW  ");
  lcd.print(__DATE__);
  delay(1000);
  lcd.clear();

  // -- pripojeni do wifi, ziskani a zobrazeni IP adresy z DHCP serveru --
  Serial.println("WiFi connecting");
  lcd.setCursor(0, 0);
  lcd.print("<IP> ");
  WiFi.persistent(false);
  WiFi.mode(WIFI_STA);
  WiFi.disconnect();
  WiFi.begin(ssid, password);
  maxRetries = 40;
  while (WiFi.status() != WL_CONNECTED) {
    ESP.wdtFeed();
    if (!maxRetries--) break;
    delay(500);
  }
  char buff[20];
  IPAddress ip = WiFi.localIP();
  sprintf(buff, "%d.%d.%d.%d", ip[0], ip[1], ip[2], ip[3]);
  Serial.println(buff);
  lcd.setCursor(0, 0);
  lcd.print(buff);

  // -- pokus o ziskani casu z NTP serveru --
  lcd.setCursor(0, 1);
  lcd.print("<NTP>");
  maxRetries = 5;
  Serial.println("NTP sync");
  while (ntpState != STBY) {
    ntptime();
    delay(500);
    ESP.wdtFeed();
    if (!maxRetries--) break;
  }
  lcd.setCursor(0, 1);
  lcd.print(myclock.sysTime);

  // -- cas pro zobrazeni uvodni obrazovky --
  delay(1000);
  lcd.clear();
}
// ----------------------------------------------------------------------------------------------




// ----------------------------------------------------------------------------------------------
// LOOP funkce
// ----------------------------------------------------------------------------------------------
void loop() {

  unsigned long currentMillis;

  // -- aktualizace casu po 1 sekunde --
  if (myclock.timeBeat()) {

    // -- pravidelna synchronizace casu s NTP --
    if (((myclock.sysTime - 10) % 900) == 0) {
      Serial.println("NTP sync");
      ntpState = SYNC;
    }

    // -- test zivota wifi s vynucenym resetem --
    if (WiFi.status() != WL_CONNECTED) {
      failedWiFi++;
      Serial.println("WiFi not connected");
    } else {
      failedWiFi = 0;
    }
    if (failedWiFi == 120) while (1);

    // -- odeslani dat pomoci HTTP API --
    if ( (myclock.sysTime % 60) == 0 ) {
      Serial.println("http get");
      if (sendviahttp()) {
        httpFails = 0;
      } else {
        httpFails++;
      }
      if (httpFails == 15) while (true);
    }

    // -- blikani LED pro orientaci, ze smycka bezi --
    if (digitalRead(LED_BUILTIN)) {
      digitalWrite(LED_BUILTIN, LOW);
    } else {
      digitalWrite(LED_BUILTIN, HIGH);
    }

    // -- zobrazeni na LCD --
    zobrazeni();

    // -- timeout na validni data od menice --
    if (fvdata[0].dataage < 1000) fvdata[0].dataage++;

  }

  // -- reset poslednich znamych hodnot od menice --
  if (fvdata[0].dataage >= 60) {
    fvdata[0].u = 0;
    fvdata[0].i = 0;
    fvdata[0].p = 0;
    fvdata[0].pwm = 0;
  }

  // -- flag zakazani prepnuti pri malem napeti FVE --
  if (fveSwitchEnable) {
    if (fvdata[0].u < 18) fveSwitchEnable = false;
  } else {
    if (fvdata[0].u > 20) fveSwitchEnable = true;
  }

  // -- zakazani / povoleni prepnuti na FVE --
  if (fveSwitchEnable) {
    pinMode(FVESWITCHENABLEPIN, INPUT);
  } else {
    pinMode(FVESWITCHENABLEPIN, OUTPUT);
  }

  // -- restart MTTP z duvodu vynulovani vyroby v noci --
  if ((myclock.dte.Hour == 23) && (myclock.dte.Minute >= 58)) {
    pinMode(FVERESTARTPIN, OUTPUT);
  } else {
    pinMode(FVERESTARTPIN, INPUT);
  }

  // -- proces synchronizace s NTP --
  ntptime();

  // -- smazani psa --
  ESP.wdtFeed();

  // -- obsluha prijatych dat ze seriove linky od MPPT regulatoru --
  while (Serial.available()) {
    parseSerial((char)Serial.read());
  }

}
// ----------------------------------------------------------------------------------------------




// ----------------------------------------------------------------------------------------------
// Obsluha prijatych dat na seriove lince od MPPT regulatoru
// ----------------------------------------------------------------------------------------------
void parseSerial(unsigned char c) {
  unsigned char k = 0;
  unsigned int val = 0;

  Serial.write(c);

  if (c == '{') {
    rxPtr = 0;
    rxCislo = 0;
    rxPtrn = 0;
  }
  if ( (rxPtr >= RXBUFSIZE) || (rxPtrn >= TMPBUFSIZE) ) return;
  rxBuffer[rxPtr++] = c;
  if ( (c >= '0') && (c <= '9') ) {
    rxTmp[rxPtrn++] = c;
    rxTmp[rxPtrn] = 0;
  }
  if ( (c == '|') || (c == '}') ) {
    val = atoi(rxTmp);
    if (rxCislo < ITEMSCOUNT) rxData[rxCislo] = val;
    rxPtrn = 0;
    rxCislo++;
  }

  if (rxCislo == ITEMSCOUNT) {
    if ((rxData[0]^rxData[1]^rxData[2]^rxData[3]^rxData[4]^rxData[5]^rxData[6]) == rxData[7]) {
      if (rxData[0] == 0) {
        fvdata[rxData[0]].dataage = 0;
        fvdata[rxData[0]].enable = rxData[1];
        fvdata[rxData[0]].i = rxData[2];
        fvdata[rxData[0]].u = rxData[3];
        fvdata[rxData[0]].p = rxData[4];
        fvdata[rxData[0]].wh = rxData[5];
        fvdata[rxData[0]].pwm = rxData[6];
      }
    }
    rxPtr = 0;
    rxCislo = 0;
    rxPtrn = 0;
  }

}
// ----------------------------------------------------------------------------------------------



// ----------------------------------------------------------------------------------------------
// Odeslani dat pres http get
// ----------------------------------------------------------------------------------------------
unsigned char sendviahttp() {
  if ((WiFi.status() == WL_CONNECTED)) {
    WiFiClient client;
    HTTPClient http;
    char sendBuffer[1000];
    sprintf(sendBuffer, "http://www.domena.cz/skript.php?cet=%d&f0_enable=%d&f0_p=%d&f0_u=%d&f0_i=%d&f0_wh=%d&f0_pwm=%d",
            myclock.sysTime,
            fvdata[0].enable,
            fvdata[0].p,
            fvdata[0].u,
            fvdata[0].i,
            fvdata[0].wh,
            fvdata[0].pwm);
    if (http.begin(client, sendBuffer)) {
      int httpCode = http.GET();
      Serial.printf("http get return code: %d\n", httpCode);
      if (httpCode > 0) {
        if (httpCode == HTTP_CODE_OK || httpCode == HTTP_CODE_MOVED_PERMANENTLY) {
          Serial.print("http get response: ");
          String payload = http.getString();
          Serial.println(payload);
          return (true);
        }
      } else {
        Serial.printf("http get failed, error: %s\n", http.errorToString(httpCode).c_str());
      }
      http.end();
    } else {
      Serial.println("http get failed, unable to connect");
    }
  }
  return (false);
}
// ----------------------------------------------------------------------------------------------



// ----------------------------------------------------------------------------------------------
// Synchronizace s NTP
// ----------------------------------------------------------------------------------------------
void ntptime() {
  switch (ntpState) {
    case SYNC:
      WiFi.hostByName(ntpServerName, timeServerIP);
      memset(packetBuffer, 0, NTP_PACKET_SIZE);
      packetBuffer[0] = 0b11100011;   // LI, Version, Mode
      packetBuffer[1] = 0;     // Stratum, or type of clock
      packetBuffer[2] = 6;     // Polling Interval
      packetBuffer[3] = 0xEC;  // Peer Clock Precision
      packetBuffer[12]  = 49;
      packetBuffer[13]  = 0x4E;
      packetBuffer[14]  = 49;
      packetBuffer[15]  = 52;
      udp.beginPacket(timeServerIP, 123); //NTP requests are to port 123
      udp.write(packetBuffer, NTP_PACKET_SIZE);
      udp.endPacket();
      ntpState = READP;
      break;
    case READP:
      if (udp.parsePacket()) {
        udp.read(packetBuffer, NTP_PACKET_SIZE); // read the packet into the buffer
        unsigned long highWord = word(packetBuffer[40], packetBuffer[41]);
        unsigned long lowWord = word(packetBuffer[42], packetBuffer[43]);
        unsigned long secsSince1900 = highWord << 16 | lowWord;
        unsigned long epoch = secsSince1900 - 2208988800UL + 3600;
        Serial.print("NTP time: "); Serial.println(epoch);
        Serial.print("NTP diff: "); Serial.println(myclock.sysTime - epoch);
        myclock.sysTime = epoch;
        ntpState = STBY;
      }
      break;
    default:
      ntpState = STBY;
      break;
  }
}
// ----------------------------------------------------------------------------------------------



// ----------------------------------------------------------------------------------------------
// Zobrazeni na znakovem LCD displeji
// ----------------------------------------------------------------------------------------------
void zobrazeni()
{
  lcd.setCursor(0, 0);
  zobrazeni_pad(fvdata[0].p);
  lcd.print(fvdata[0].p);
  lcd.print("W ");

  lcd.setCursor(6, 0);
  zobrazeni_pad(fvdata[0].u);
  lcd.print(fvdata[0].u);
  lcd.print("V ");

  lcd.setCursor(11, 0);
  zobrazeni_pad(fvdata[0].pwm);
  lcd.print(fvdata[0].pwm);
  lcd.print("%");

  lcd.setCursor(11, 1);
  char timestr[10];
  sprintf(timestr, "%02d:%02d", myclock.dte.Hour, myclock.dte.Minute);
  lcd.print(timestr);

  lcd.setCursor(0, 1);
  if (fveSwitchEnable) {
    lcd.print("SE");
  } else {
    lcd.print("SD");
  }

  lcd.setCursor(3, 1);
  if (fvdata[0].dataage > 2) {
    lcd.print("offline");
  } else {
    lcd.print("online ");
  }

}

// -- zajisti zarovnani cisla doprava --
void zobrazeni_pad(unsigned int value) {
  if (value < 1000) lcd.print(" ");
  if (value < 100)  lcd.print(" ");
  if (value < 10)   lcd.print(" ");
}

// ----------------------------------------------------------------------------------------------
